/*
 * Copyright (c) 2004 Iwata <iwata@quasiquote.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <dlfcn.h>
#include <gauche/extend.h>
#include "lazy-ffi.h"

static void print_lazy(ScmObj, ScmPort *, ScmWriteContext *);
SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_LazyClass, print_lazy);

static void
print_lazy(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
	Scm_Printf(out, "#<lazy-ffi %S>", SCM_LAZY_FFI(obj)->name);
}


char lazy_ffi_strerr[1024];

ScmObj
Scm_LazyFfiP(ScmObj Obj)
{
	if (SCM_LAZY_FFI_P(Obj))
		return SCM_TRUE;
	return SCM_FALSE;
}
	    
ScmObj
Scm_LazyStrerr()
{
	return Scm_MakeString(lazy_ffi_strerr, strlen(lazy_ffi_strerr), -1, SCM_MAKSTR_COPYING);
}

void*
open_ffi(char *name)
{
	const char *s;
	void *handle = dlopen(name, RTLD_LAZY);
	if ((s = dlerror()) != NULL) {
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "dlopen('%s'): %s", name, s);
		return NULL;
	}
	return handle;
}

void
close_ffi(void *handle)
{
	dlclose(handle);
}

ScmObj
Scm_LazyOpen(ScmString *name)
{
	void *handle;
	ScmObj obj;

	GC_finalization_proc ofn; GC_PTR ocd;
	ScmLazy *z = SCM_NEW(ScmLazy);

	SCM_SET_CLASS(z, SCM_CLASS_LAZY);

	GC_REGISTER_FINALIZER(z, finalize_lazy, NULL, &ofn, &ocd);
	z->name = SCM_OBJ(name);

	handle = open_ffi(Scm_GetString(name));
	if (!handle)
		return SCM_FALSE;

	z->handle = handle;

	return SCM_OBJ(z);
}

ScmObj
Scm_LazyClose(ScmLazy *z)
{
	if (!z->handle) {
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "dynamic library already closed");
		return SCM_FALSE;
	}
	close_ffi(z->handle);
	z->handle = NULL;

	return SCM_UNDEFINED;
}

void*
find_function(void *handle, char *symbol)
{
	const char *s;
	void *fun;

	fun = dlsym(handle, symbol);
	if ((s = dlerror()) != NULL) {
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "dlsym('%s'): %s", symbol, s);
		return NULL;
	}
	return fun;
}

/*
 Copyright (c) 2000-2005, Felix L. Winkelmann
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 conditions are met:

   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
     disclaimer.
   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
     disclaimer in the documentation and/or other materials provided with the distribution.
   Neither the name of the author nor the names of its contributors may be used to endorse or promote
     products derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
 CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

 Send bugs, suggestions and ideas to:

 felix@call-with-current-continuation.org

 Felix L. Winkelmann
 Unter den Gleichen 1
 37130 Gleichen
 Germany
*/
static ScmObj
intern_Scm_LazyCall(ScmLazy *z, int rtype, void *fun, ScmObj args, ScmObj atypelist)
{
	int i;
	int n = Scm_Length(args);
	ffi_status s;
	ffi_cif cif;
	void **avalues = alloca(n * sizeof(void *));
	ffi_type **atypes = alloca(n * sizeof(ffi_type *));
	void **avptr = avalues;
	ffi_type **atptr = atypes;
	ffi_type *rt;
	void *ptr, *resultb = alloca(8);
	ScmObj cp, ap;

	switch(rtype) {
	case LAZY_VOID_TYPE: rt = &ffi_type_void; break;
	case LAZY_INT_TYPE: rt = &ffi_type_sint; break;
	case LAZY_DOUBLE_TYPE: rt = &ffi_type_double; break;
	case LAZY_SCHEME_PTR_TYPE:
	case LAZY_SCHEME_TYPE:
	case LAZY_STRING_TYPE:
	case LAZY_PTR_TYPE: rt = &ffi_type_pointer; break;
	case LAZY_CHAR_TYPE: rt = &ffi_type_uint8; break;
	default:
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "unknown output type: '%d'", rtype);
		return SCM_FALSE;
	}

	for(i = 0, cp = args, ap = atypelist; i < n; ++i, cp = SCM_CDR(cp), ap = SCM_CDR(ap)) {
		ScmObj arg   = SCM_CAR(cp);
		ScmObj atype = SCM_CAR(ap);

		if (!SCM_INTEGERP(atype)) {
			snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "atype must be number", rtype);
			return SCM_FALSE;
		}
		switch (Scm_GetInteger(atype)) {
		case LAZY_INT_TYPE:
			ptr = alloca(sizeof(int));
			*((int *)ptr) = Scm_GetInteger(arg);
			*(atptr++) = &ffi_type_sint;
			*(avptr++) = ptr;
			break;

		case LAZY_CHAR_TYPE:
			ptr = alloca(sizeof(int));
			*((char *)ptr) = SCM_CHAR_VALUE(arg);
			*(atptr++) = &ffi_type_uint8;
			*(avptr++) = ptr;
			break;

		case LAZY_DOUBLE_TYPE:
			ptr = alloca(sizeof(double));
			*((double *)ptr) = Scm_GetDouble(arg);
			*(atptr++) = &ffi_type_double;
			*(avptr++) = ptr;
			break;

		case LAZY_SCHEME_PTR_TYPE:
			if (SCM_IMMEDIATEP(arg)) {
				snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr
), "immidiate type: '%d'", rtype);
				return SCM_FALSE;
			}
			/* fall through... */

		case LAZY_STRING_TYPE:
			ptr = alloca(sizeof(void *));
			*(atptr++) = &ffi_type_pointer;
			*(avptr++) = ptr;
			*((void **)ptr) = (void *)Scm_GetString(arg);
			break;

		case LAZY_PTR_TYPE:
			ptr = alloca(sizeof(void *));
			*(atptr++) = &ffi_type_pointer;
			*(avptr++) = ptr;
			*((void **)ptr) = (void *)SCM_C_POINTER(arg)->ptr;
			break;

		case LAZY_SCHEME_TYPE:
			*(atptr++) = &ffi_type_pointer;
			*(avptr++) = (void *)arg;
			break;

		default:
			snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "unknown input type: '%d'", rtype);
			return SCM_FALSE;
		}
	}

	s = ffi_prep_cif(&cif, FFI_DEFAULT_ABI, n, rt, atypes);

	if(s != FFI_OK) {
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "ffi_prep_cif: '%d'", s);
		return SCM_FALSE;
	}

	ffi_call(&cif, fun, resultb, avalues);

	switch(rtype) {
	case LAZY_INT_TYPE:
		return Scm_MakeInteger(*((int *)resultb));
	case LAZY_DOUBLE_TYPE:
		return Scm_MakeFlonum(*((double *)resultb));
	case LAZY_STRING_TYPE:
		return SCM_MAKE_STR(*((char **)resultb));
	case LAZY_PTR_TYPE:
		return Scm_LazyMakeCpointer(*((int *)resultb));
	case LAZY_CHAR_TYPE:
		return SCM_MAKE_CHAR(*((char *)resultb));
	case LAZY_SCHEME_TYPE:
		return *((ScmObj *)resultb);
	case LAZY_VOID_TYPE:
		return SCM_UNDEFINED;
	default:
		return SCM_FALSE;
	}
	return SCM_FALSE;
}

ScmObj
Scm_LazyCall(ScmLazy *z, int rtype, ScmString *name, ScmObj args, ScmObj atypelist)
{
	void *fun;

	if (z->handle == NULL) {
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "dynamic library not initialised");
		return SCM_FALSE;
	}
	fun = find_function(z->handle, Scm_GetString(name));
	if (!fun)
		return SCM_FALSE;

	if (Scm_Length(args) != Scm_Length(atypelist)) {
		snprintf(lazy_ffi_strerr, sizeof(lazy_ffi_strerr), "argument length mismatch", rtype);
		return SCM_FALSE;
	}

	return intern_Scm_LazyCall(z, rtype, fun, args, atypelist);
}

void
finalize_lazy(GC_PTR obj, GC_PTR data)
{
	ScmLazy *lazy = SCM_LAZY_FFI(obj);
	if (lazy->handle) {
		close_ffi(lazy->handle);
		lazy->handle = NULL;
	}
}

extern void Scm_Init_lib_lazy_ffi(ScmModule*);

void
Scm_Init_lazy_ffi(void)
{
	ScmModule *mod;
	SCM_INIT_EXTENSION(lazyffi);
	mod = SCM_MODULE(SCM_FIND_MODULE("lazy-ffi", TRUE));
	Scm_InitStaticClass(&Scm_LazyClass, "<lazy-ffi>", mod, NULL, 0);
	Scm_InitStaticClass(&Scm_LazyClass, "<c-pointer>", mod, NULL, 0);
	Scm_Init_lib_lazy_ffi(mod);
}


